'''
Image
=====

The :class:`Image` widget is used to display an image::

Example in python::

    wimg = Image(source='mylogo.png')

Kv Example::

    Image:
        source: 'mylogo.png'
        size: self.texture_size


Asynchronous Loading
--------------------

To load an image asynchronously (for example from an external webserver), use
the :class:`AsyncImage` subclass::

    aimg = AsyncImage(source='http://mywebsite.com/logo.png')

This can be useful as it prevents your application from waiting until the image
is loaded. If you want to display large images or retrieve them from URL's,
using :class:`AsyncImage` will allow these resources to be retrieved on a
background thread without blocking your application.

Alignment
---------

By default, the image is centered inside the widget bounding box.

Adjustment
----------

To control how the image should be adjusted to fit inside the widget box, you
should use the :attr:`~kivy.uix.image.Image.fit_mode` property. Available
options include:

- ``"scale-down"``: maintains aspect ratio without stretching.
- ``"fill"``: stretches to fill widget, may cause distortion.
- ``"contain"``: maintains aspect ratio and resizes to fit inside widget.
- ``"cover"``: maintains aspect ratio and stretches to fill widget, may clip
image.

For more details, refer to the :attr:`~kivy.uix.image.Image.fit_mode`.


You can also inherit from Image and create your own style. For example, if you
want your image to be greater than the size of your widget, you could do::

    class FullImage(Image):
        pass

And in your kivy language file::

    <-FullImage>:
        canvas:
            Color:
                rgb: (1, 1, 1)
            Rectangle:
                texture: self.texture
                size: self.width + 20, self.height + 20
                pos: self.x - 10, self.y - 10

'''
__all__ = ('Image', 'AsyncImage')

from kivy.uix.widget import Widget
from kivy.core.image import Image as CoreImage
from kivy.resources import resource_find
from kivy.properties import (
    StringProperty,
    ObjectProperty,
    ListProperty,
    AliasProperty,
    BooleanProperty,
    NumericProperty,
    ColorProperty,
    OptionProperty
)
from kivy.logger import Logger

# delayed imports
Loader = None


class Image(Widget):
    '''Image class, see module documentation for more information.'''

    source = StringProperty(None)
    '''Filename / source of your image.

    :attr:`source` is a :class:`~kivy.properties.StringProperty` and
    defaults to None.
    '''

    texture = ObjectProperty(None, allownone=True)
    '''Texture object of the image. The texture represents the original, loaded
    image texture. It is stretched and positioned during rendering according to
    the :attr:`fit_mode` property.

    Depending of the texture creation, the value will be a
    :class:`~kivy.graphics.texture.Texture` or a
    :class:`~kivy.graphics.texture.TextureRegion` object.

    :attr:`texture` is an :class:`~kivy.properties.ObjectProperty` and defaults
    to None.
    '''

    texture_size = ListProperty([0, 0])
    '''Texture size of the image. This represents the original, loaded image
    texture size.

    .. warning::

        The texture size is set after the texture property. So if you listen to
        the change on :attr:`texture`, the property texture_size will not be
        up-to-date. Use self.texture.size instead.
    '''

    def get_image_ratio(self):
        if self.texture:
            return self.texture.width / float(self.texture.height)
        return 1.0

    mipmap = BooleanProperty(False)
    '''Indicate if you want OpenGL mipmapping to be applied to the texture.
    Read :ref:`mipmap` for more information.

    .. versionadded:: 1.0.7

    :attr:`mipmap` is a :class:`~kivy.properties.BooleanProperty` and defaults
    to False.
    '''

    image_ratio = AliasProperty(get_image_ratio, bind=('texture',), cache=True)
    '''Ratio of the image (width / float(height).

    :attr:`image_ratio` is an :class:`~kivy.properties.AliasProperty` and is
    read-only.
    '''

    color = ColorProperty([1, 1, 1, 1])
    '''Image color, in the format (r, g, b, a). This attribute can be used to
    'tint' an image. Be careful: if the source image is not gray/white, the
    color will not really work as expected.

    .. versionadded:: 1.0.6

    :attr:`color` is a :class:`~kivy.properties.ColorProperty` and defaults to
    [1, 1, 1, 1].

    .. versionchanged:: 2.0.0
        Changed from :class:`~kivy.properties.ListProperty` to
        :class:`~kivy.properties.ColorProperty`.
    '''

    allow_stretch = BooleanProperty(False, deprecated=True)
    '''If True, the normalized image size will be maximized to fit in the image
    box. Otherwise, if the box is too tall, the image will not be
    stretched more than 1:1 pixels.

    .. versionadded:: 1.0.7

    .. deprecated:: 2.2.0
        :attr:`allow_stretch` have been deprecated. Please use `fit_mode`
        instead.

    :attr:`allow_stretch` is a :class:`~kivy.properties.BooleanProperty` and
    defaults to False.
    '''

    keep_ratio = BooleanProperty(True, deprecated=True)
    '''If False along with allow_stretch being True, the normalized image
    size will be maximized to fit in the image box and ignores the aspect
    ratio of the image.
    Otherwise, if the box is too tall, the image will not be stretched more
    than 1:1 pixels.

    .. versionadded:: 1.0.8

    .. deprecated:: 2.2.0
        :attr:`keep_ratio` have been deprecated. Please use `fit_mode`
        instead.

    :attr:`keep_ratio` is a :class:`~kivy.properties.BooleanProperty` and
    defaults to True.
    '''

    fit_mode = OptionProperty(
        "scale-down", options=["scale-down", "fill", "contain", "cover"]
    )
    '''If the size of the image is different than the size of the widget,
    determine how the image should be resized to fit inside the widget box.

    Available options:

    - ``"scale-down"``: the image will be scaled down to fit inside the widget
    box, **maintaining its aspect ratio and without stretching**. If the size
    of the image is smaller than the widget, it will be displayed at its
    original size. If the image has a different aspect ratio than the widget,
    there will be blank areas on the widget box.

    - ``"fill"``: the image is stretched to fill the widget, **regardless of
    its aspect ratio or dimensions**. If the image has a different aspect ratio
    than the widget, this option can lead to distortion of the image.

    - ``"contain"``: the image is resized to fit inside the widget box,
    **maintaining its aspect ratio**. If the image size is larger than the
    widget size, the behavior will be similar to ``"scale-down"``. However, if
    the size of the image size is smaller than the widget size, unlike
    ``"scale-down``, the image will be resized to fit inside the widget.
    If the image has a different aspect ratio than the widget, there will be
    blank areas on the widget box.

    - ``"cover"``: the image will be stretched horizontally or vertically to
    fill the widget box, **maintaining its aspect ratio**. If the image has a
    different aspect ratio than the widget, then the image will be clipped to
    fit.

    :attr:`fit_mode` is a :class:`~kivy.properties.OptionProperty` and
    defaults to ``"scale-down"``.
    '''

    keep_data = BooleanProperty(False)
    '''If True, the underlying _coreimage will store the raw image data.
    This is useful when performing pixel based collision detection.

    .. versionadded:: 1.3.0

    :attr:`keep_data` is a :class:`~kivy.properties.BooleanProperty` and
    defaults to False.
    '''

    anim_delay = NumericProperty(0.25)
    '''Delay the animation if the image is sequenced (like an animated gif).
    If anim_delay is set to -1, the animation will be stopped.

    .. versionadded:: 1.0.8

    :attr:`anim_delay` is a :class:`~kivy.properties.NumericProperty` and
    defaults to 0.25 (4 FPS).
    '''

    anim_loop = NumericProperty(0)
    '''Number of loops to play then stop animating. 0 means keep animating.

    .. versionadded:: 1.9.0

    :attr:`anim_loop` is a :class:`~kivy.properties.NumericProperty` and
    defaults to 0.
    '''

    nocache = BooleanProperty(False)
    '''If this property is set True, the image will not be added to the
    internal cache. The cache will simply ignore any calls trying to
    append the core image.

    .. versionadded:: 1.6.0

    :attr:`nocache` is a :class:`~kivy.properties.BooleanProperty` and defaults
    to False.
    '''

    def get_norm_image_size(self):
        if not self.texture:
            return list(self.size)

        ratio = self.image_ratio
        w, h = self.size
        tw, th = self.texture.size

        if self.fit_mode == "cover":
            widget_ratio = w / max(1, h)
            if widget_ratio > ratio:
                return [w, (w * th) / tw]
            else:
                return [(h * tw) / th, h]
        elif self.fit_mode == "fill":
            return [w, h]
        elif self.fit_mode == "contain":
            iw = w
        else:
            iw = min(w, tw)

        # calculate the appropriate height
        ih = iw / ratio
        # if the height is too higher, take the height of the container
        # and calculate appropriate width. no need to test further. :)
        if ih > h:
            if self.fit_mode == "contain":
                ih = h
            else:
                ih = min(h, th)
            iw = ih * ratio
        return [iw, ih]

    norm_image_size = AliasProperty(
        get_norm_image_size,
        bind=(
            'texture',
            'size',
            'image_ratio',
            'fit_mode',
        ),
        cache=True,
    )
    '''Normalized image size within the widget box.

    This size will always fit the widget size and will preserve the image
    ratio.

    :attr:`norm_image_size` is an :class:`~kivy.properties.AliasProperty` and
    is read-only.
    '''

    def __init__(self, **kwargs):
        self._coreimage = None
        self._loops = 0
        update = self.texture_update
        fbind = self.fbind
        fbind('source', update)
        fbind('mipmap', update)

        # NOTE: Compatibility code due to deprecated properties.
        fbind('keep_ratio', self._update_fit_mode)
        fbind('allow_stretch', self._update_fit_mode)
        super().__init__(**kwargs)

    def _update_fit_mode(self, *args):
        keep_ratio = self.keep_ratio
        allow_stretch = self.allow_stretch
        if (
            not keep_ratio and not allow_stretch
            or keep_ratio and not allow_stretch
        ):
            self.fit_mode = "scale-down"
        elif not keep_ratio and allow_stretch:
            self.fit_mode = "fill"
        elif keep_ratio and allow_stretch:
            self.fit_mode = "contain"

    def texture_update(self, *largs):
        self.set_texture_from_resource(self.source)

    def set_texture_from_resource(self, resource):
        if not resource:
            self._clear_core_image()
            return
        source = resource_find(resource)
        if not source:
            Logger.error('Image: Not found <%s>' % resource)
            self._clear_core_image()
            return
        if self._coreimage:
            self._coreimage.unbind(on_texture=self._on_tex_change)
        try:
            self._coreimage = image = CoreImage(
                source,
                mipmap=self.mipmap,
                anim_delay=self.anim_delay,
                keep_data=self.keep_data,
                nocache=self.nocache
            )
        except Exception:
            Logger.error('Image: Error loading <%s>' % resource)
            self._clear_core_image()
            image = self._coreimage
        if image:
            image.bind(on_texture=self._on_tex_change)
            self.texture = image.texture

    def on_anim_delay(self, instance, value):
        if self._coreimage is None:
            return
        self._coreimage.anim_delay = value
        if value < 0:
            self._coreimage.anim_reset(False)

    def on_texture(self, instance, value):
        self.texture_size = value.size if value else [0, 0]

    def _clear_core_image(self):
        if self._coreimage:
            self._coreimage.unbind(on_texture=self._on_tex_change)
        self.texture = None
        self._coreimage = None
        self._loops = 0

    def _on_tex_change(self, *largs):
        # update texture from core image
        self.texture = self._coreimage.texture
        ci = self._coreimage
        if self.anim_loop and ci._anim_index == len(ci._image.textures) - 1:
            self._loops += 1
            if self.anim_loop == self._loops:
                ci.anim_reset(False)
                self._loops = 0

    def reload(self):
        '''Reload image from disk. This facilitates re-loading of
        images from disk in case the image content changes.

        .. versionadded:: 1.3.0

        Usage::

            im = Image(source = '1.jpg')
            # -- do something --
            im.reload()
            # image will be re-loaded from disk

        '''
        self.remove_from_cache()
        old_source = self.source
        self.source = ''
        self.source = old_source

    def remove_from_cache(self):
        '''Remove image from cache.

        .. versionadded:: 2.0.0
        '''
        if self._coreimage:
            self._coreimage.remove_from_cache()

    def on_nocache(self, *args):
        if self.nocache:
            self.remove_from_cache()
            if self._coreimage:
                self._coreimage._nocache = True


class AsyncImage(Image):
    '''Asynchronous Image class. See the module documentation for more
    information.

    .. note::

        The AsyncImage is a specialized form of the Image class. You may
        want to refer to the :mod:`~kivy.loader` documentation and in
        particular, the :class:`~kivy.loader.ProxyImage` for more detail
        on how to handle events around asynchronous image loading.

    .. note::

        AsyncImage currently does not support properties
        :attr:`anim_loop` and :attr:`mipmap` and setting those properties will
        have no effect.
    '''

    __events__ = ('on_error', 'on_load')

    def __init__(self, **kwargs):
        self._found_source = None
        self._coreimage = None
        global Loader
        if not Loader:
            from kivy.loader import Loader
        self.fbind('source', self._load_source)
        super().__init__(**kwargs)

    def _load_source(self, *args):
        source = self.source
        if not source:
            self._clear_core_image()
            return
        if not self.is_uri(source):
            source = resource_find(source)
            if not source:
                Logger.error('AsyncImage: Not found <%s>' % self.source)
                self._clear_core_image()
                return
        self._found_source = source
        self._coreimage = image = Loader.image(
            source,
            nocache=self.nocache,
            mipmap=self.mipmap,
            anim_delay=self.anim_delay
        )
        image.bind(
            on_load=self._on_source_load,
            on_error=self._on_source_error,
            on_texture=self._on_tex_change
        )
        self.texture = image.texture

    def _on_source_load(self, value):
        image = self._coreimage.image
        if not image:
            return
        self.texture = image.texture
        self.dispatch('on_load')

    def _on_source_error(self, instance, error=None):
        self.dispatch('on_error', error)

    def on_error(self, error):
        pass

    def on_load(self, *args):
        pass

    def is_uri(self, filename):
        proto = filename.split('://', 1)[0]
        return proto in ('http', 'https', 'ftp', 'smb')

    def _clear_core_image(self):
        if self._coreimage:
            self._coreimage.unbind(on_load=self._on_source_load)
        super()._clear_core_image()
        self._found_source = None

    def _on_tex_change(self, *largs):
        if self._coreimage:
            self.texture = self._coreimage.texture

    def texture_update(self, *largs):
        pass

    def remove_from_cache(self):
        if self._found_source:
            Loader.remove_from_cache(self._found_source)
        super().remove_from_cache()



"""
Logger tests
============
"""

import logging
import os
import pathlib
import sys
import time

import pytest

# Used to determine which tests must be skipped
LOG_MODE = os.environ.get("KIVY_LOG_MODE", "KIVY")


@pytest.fixture
def file_handler():
    # restores handler to original state
    from kivy.config import Config

    log_dir = Config.get("kivy", "log_dir")
    log_maxfiles = Config.get("kivy", "log_maxfiles")

    try:
        yield None
    finally:
        Config.set("kivy", "log_dir", log_dir)
        Config.set("kivy", "log_maxfiles", log_maxfiles)


@pytest.mark.parametrize("n", [0, 1, 5])
def test_purge_logs(tmp_path, file_handler, n):
    from kivy.config import Config
    from kivy.logger import FileHandler

    Config.set("kivy", "log_dir", str(tmp_path))
    Config.set("kivy", "log_maxfiles", n)

    # create the default file first so it gets deleted so names match
    handler = FileHandler()
    handler._configure()
    open_file = pathlib.Path(handler.filename).name
    # wait a little so the timestamps are different for different files
    time.sleep(0.05)

    names = [f"log_{i}.txt" for i in range(n + 2)]
    for name in names:
        p = tmp_path / name
        p.write_text("some data")
        time.sleep(0.05)

    handler.purge_logs()

    # files that should have remained after purge
    expected_names = list(reversed(names))[:n]
    files = {f.name for f in tmp_path.iterdir()}
    if open_file in files:
        # one of the remaining files is the current open log, remove it
        files.remove(open_file)
        if len(expected_names) == len(files) + 1:
            # the open log may or may not have been counted in the remaining
            # files, remove one from expected to match removed open file
            expected_names = expected_names[:-1]

    assert set(expected_names) == files


@pytest.mark.xfail  # Issue #7986 not yet fixed.
def test_logger_history_size():
    from kivy.logger import Logger, LoggerHistory

    for x in range(200):
        Logger.info(x)
    assert len(LoggerHistory.history) == 100, "Wrong size: %s" % len(
        LoggerHistory.history
    )


@pytest.mark.skipif(
    LOG_MODE != "KIVY",
    reason="Requires KIVY_LOG_MODE==KIVY to run.",
)
def test_trace_level():
    """Kivy logger defines a custom level of Trace."""
    from kivy.logger import Logger, LOG_LEVELS, LoggerHistory
    import logging

    Logger.setLevel(9)

    # Try different ways to trigger a trace:
    Logger.trace("test: This is trace message 1")
    logging.log(logging.TRACE, "test: This is trace message 2")
    Logger.log(LOG_LEVELS["trace"], "test: This is trace message 3")
    # Not supported:
    # logging.trace('test: This is trace message 4')

    last_log_records = LoggerHistory.history[:3]
    assert all(log_record.levelno == 9 for log_record in last_log_records), [
        log_record.levelno for log_record in last_log_records
    ]


@pytest.mark.skipif(
    LOG_MODE == "PYTHON",
    reason="Requires KIVY_LOG_MODE!=PYTHON to run.",
)
def test_trace_level_has_level_name():
    from kivy.logger import Logger, LoggerHistory

    Logger.setLevel(9)
    Logger.trace("test: This is trace message 1")
    assert LoggerHistory.history[0].levelname == "TRACE"


def test_logging_does_not_deep_copy():
    # If the Logger does a deep copy of an uncopyable
    # data structure, it will fail. See issues #7585 and #7528.

    import threading
    from kivy.logger import Logger

    class UncopyableDatastructure:
        def __init__(self, name):
            self._lock = threading.Lock()
            self._name = name

        def __str__(self):
            return "UncopyableDatastructure(name=%r)" % self._name

    s = UncopyableDatastructure("Uncopyable")
    Logger.error("The value of s is %s", s)


def configured_string_logging(unique_code, formatter=None):
    """
    Helper function provides logger configured to write to log_output.
    """
    from io import StringIO

    log_output = StringIO()

    handler = logging.StreamHandler(stream=log_output)
    if formatter:
        handler.setFormatter(formatter)

    logger = logging.getLogger("tests.%s" % unique_code)

    # Do not escalate to root/Kivy loggers.
    logger.setLevel(9)  # Catch everything
    logger.propagate = False
    assert not logger.hasHandlers(), "Must use unique code between tests."

    logger.addHandler(handler)

    return logger, log_output


def test_colonsplittinglogrecord_with_colon():
    from kivy.logger import ColonSplittingLogRecord

    originallogrecord = logging.LogRecord(
        name="kivy.test",
        level=logging.DEBUG,
        pathname="test.py",
        lineno=1,
        msg="Part1: Part2: Part 3",
        args=("args",),
        exc_info=None,
        func="test_colon_splitting",
        sinfo=None,
    )
    # Just making sure we know what it looks like before.
    assert (
        str(originallogrecord)
        == '<LogRecord: kivy.test, 10, test.py, 1, "Part1: Part2: Part 3">'
    )
    shimmedlogrecord = ColonSplittingLogRecord(originallogrecord)
    assert (
        str(shimmedlogrecord) == "<LogRecord: kivy.test, 10, test.py, 1, "
        '"[Part1       ] Part2: Part 3">'
    )


def test_colonsplittinglogrecord_without_colon():
    from kivy.logger import ColonSplittingLogRecord

    originallogrecord = logging.LogRecord(
        name="kivy.test",
        level=logging.DEBUG,
        pathname="test.py",
        lineno=1,
        msg="Part1 Part2 Part 3",
        args=("args",),
        exc_info=None,
        func="test_colon_splitting",
        sinfo=None,
    )
    shimmedlogrecord = ColonSplittingLogRecord(originallogrecord)
    # No colons means no change.
    assert str(originallogrecord) == str(shimmedlogrecord)

    # Try a non-string (Issue #7984)
    originallogrecord = logging.LogRecord(
        name="kivy.test",
        level=logging.DEBUG,
        pathname="test.py",
        lineno=1,
        msg=1,
        args=None,
        exc_info=None,
        func="test_colon_splitting",
        sinfo=None,
    )
    shimmedlogrecord = ColonSplittingLogRecord(originallogrecord)
    # No colons means no change.
    assert str(originallogrecord) == str(shimmedlogrecord)


def test_uncoloredlogrecord_without_markup():
    from kivy.logger import UncoloredLogRecord

    originallogrecord = logging.LogRecord(
        name="kivy.test",
        level=logging.DEBUG,
        pathname="test.py",
        lineno=1,
        msg="Part1: Part2 Part 3",
        args=("args",),
        exc_info=None,
        func="test_colon_splitting",
        sinfo=None,
    )
    shimmedlogrecord = UncoloredLogRecord(originallogrecord)
    # No markup means no change.
    assert str(originallogrecord) == str(shimmedlogrecord)

    # Try a non-string (Issue #7984)
    originallogrecord = logging.LogRecord(
        name="kivy.test",
        level=logging.DEBUG,
        pathname="test.py",
        lineno=1,
        msg=1,
        args=None,
        exc_info=None,
        func="test_colon_splitting",
        sinfo=None,
    )
    shimmedlogrecord = UncoloredLogRecord(originallogrecord)
    assert str(originallogrecord) == str(shimmedlogrecord)


def test_uncoloredlogrecord_with_markup():
    from kivy.logger import UncoloredLogRecord

    originallogrecord = logging.LogRecord(
        name="kivy.test",
        level=logging.DEBUG,
        pathname="test.py",
        lineno=1,
        msg="Part1: $BOLDPart2$RESET Part 3",
        args=("args",),
        exc_info=None,
        func="test_colon_splitting",
        sinfo=None,
    )
    shimmedlogrecord = UncoloredLogRecord(originallogrecord)
    # No markup means no change.
    assert (
        str(shimmedlogrecord)
        == '<LogRecord: kivy.test, 10, test.py, 1, "Part1: Part2 Part 3">'
    )


def test_coloredlogrecord_without_markup():
    from kivy.logger import ColoredLogRecord

    originallogrecord = logging.LogRecord(
        name="kivy.test",
        level=logging.DEBUG,
        pathname="test.py",
        lineno=1,
        msg="Part1: Part2 Part 3",
        args=("args",),
        exc_info=None,
        func="test_colon_splitting",
        sinfo=None,
    )
    shimmedlogrecord = ColoredLogRecord(originallogrecord)
    # The str() looks the same, because it doesn't include levelname.
    assert str(originallogrecord) == str(shimmedlogrecord)
    # But there is a change in the levelname
    assert originallogrecord.levelname != shimmedlogrecord.levelname
    assert shimmedlogrecord.levelname == "\x1b[1;36mDEBUG\x1b[0m"

    # Try a non-string (Issue #7984)
    originallogrecord = logging.LogRecord(
        name="kivy.test",
        level=logging.DEBUG,
        pathname="test.py",
        lineno=1,
        msg=1,
        args=None,
        exc_info=None,
        func="test_colon_splitting",
        sinfo=None,
    )
    shimmedlogrecord = ColoredLogRecord(originallogrecord)
    assert str(originallogrecord) == str(shimmedlogrecord)


def test_coloredlogrecord_with_markup():
    from kivy.logger import ColoredLogRecord

    originallogrecord = logging.LogRecord(
        name="kivy.test",
        level=logging.INFO,
        pathname="test.py",
        lineno=1,
        msg="Part1: $BOLDPart2$RESET Part 3",
        args=("args",),
        exc_info=None,
        func="test_colon_splitting",
        sinfo=None,
    )
    shimmedlogrecord = ColoredLogRecord(originallogrecord)
    # Bolding has been added to message.
    assert (
        str(shimmedlogrecord) == "<LogRecord: kivy.test, 20, test.py, 1, "
        '"Part1: \x1b[1mPart2\x1b[0m Part 3">'
    )
    # And there is a change in the levelname
    assert originallogrecord.levelname != shimmedlogrecord.levelname
    assert shimmedlogrecord.levelname == "\x1b[1;32mINFO\x1b[0m"


def test_kivyformatter_colon_no_color():
    from kivy.logger import KivyFormatter

    formatter = KivyFormatter("[%(levelname)-7s] %(message)s", use_color=False)
    logger, log_output = configured_string_logging("1", formatter)
    logger.info("Fancy: $BOLDmess$RESETage")
    assert log_output.getvalue() == "[INFO   ] [Fancy       ] message\n"


def test_kivyformatter_colon_color():
    from kivy.logger import KivyFormatter

    formatter = KivyFormatter("[%(levelname)-18s] %(message)s", use_color=True)

    logger, log_output = configured_string_logging("2", formatter)
    logger.info("Fancy: $BOLDmess$RESETage")
    assert (
        log_output.getvalue()
        == "[\x1b[1;32mINFO\x1b[0m   ] [Fancy       ] \x1b[1mmess\x1b[0mage\n"
    )


@pytest.mark.logmodepython
@pytest.mark.logmodemixed
@pytest.mark.skipif(
    LOG_MODE == "KIVY",
    reason="Requires KIVY_LOG_MODE!=KIVY to run.",
)
def test_kivy_log_mode_marker_on():
    """
    This is a test of the pytest markers.
    This should only be invoked if the environment variable is appropriately set
    (before pytest is run).

    Also, tests that kivy.logger paid attention to the environment variable
    """
    assert logging.root.level != 0, "Root logger was modified"


@pytest.mark.skipif(
    LOG_MODE != "KIVY",
    reason="Requires KIVY_LOG_MODE==KIVY to run.",
)
def test_kivy_log_mode_marker_off():
    """
    This is a test of the pytest markers.
    This should only be invoked if the environment variable is properly set
    (before pytest is run).

    Also, tests that kivy.logger paid attention to the environment variable
    """
    assert logging.root.level == 0, "Root logger was not modified"


# These tests are an overly-dramatic simulation of a third-party library's
# logging infrastructure.
#
# A handler is attached to the root-logger, which takes action when a
# critical message is logged.


class NuclearReactorMonitoringHandler(logging.Handler):
    NUCLEAR_REACTOR_STATUS = "Nominal"

    def __init__(self):
        super().__init__(level=logging.CRITICAL)

    def emit(self, log_record):
        # The fact that this was called means a critical log record
        # was created.
        sys.stderr.write("Please proceed immediately to the nearest exit.\n")
        sys.stderr.flush()
        self.NUCLEAR_REACTOR_STATUS = "Evacuated"


def simulate_evacuation():
    logging.getLogger().addHandler(logging.StreamHandler())

    handler = NuclearReactorMonitoringHandler()
    logging.getLogger().addHandler(handler)

    nuclear_core_logger = logging.getLogger("powerstation.core")
    nuclear_core_logger.info("Core temperature nominal")
    assert handler.NUCLEAR_REACTOR_STATUS == "Nominal"
    nuclear_core_logger.critical("Radioactive gas leak")
    assert handler.NUCLEAR_REACTOR_STATUS == "Evacuated"


@pytest.mark.logmodepython
@pytest.mark.logmodemixed
def test_third_party_handlers_work():
    # This should work under any mode.
    simulate_evacuation()


def are_regular_logs_handled():
    from kivy.logger import LoggerHistory

    LoggerHistory.clear_history()
    logging.getLogger("test").info(1)
    return bool(LoggerHistory.history)


def are_kivy_logger_logs_handled():
    from kivy.logger import LoggerHistory, Logger

    LoggerHistory.clear_history()
    Logger.info(1)
    return bool(LoggerHistory.history)


def is_stderr_output_handled():
    from kivy.logger import LoggerHistory

    LoggerHistory.clear_history()
    sys.stderr.write("Test output to stderr\n")
    sys.stderr.flush()
    return bool(LoggerHistory.history)


@pytest.mark.skipif(
    LOG_MODE != "KIVY",
    reason="Requires KIVY_LOG_MODE==KIVY to run.",
)
def test_kivy_mode_handlers():
    assert are_regular_logs_handled()
    assert are_kivy_logger_logs_handled()
    # This line doesn't work because of pytest's handling of stderr.
    # Runs fine as stand-alone code.
    # assert is_stderr_output_handled()


@pytest.mark.logmodepython
@pytest.mark.skipif(
    LOG_MODE != "PYTHON",
    reason="Requires KIVY_LOG_MODE==PYTHON to run.",
)
def test_python_mode_handlers():
    assert not are_regular_logs_handled()
    assert not are_kivy_logger_logs_handled()
    assert not is_stderr_output_handled()


@pytest.mark.logmodemixed
@pytest.mark.skipif(
    LOG_MODE != "MIXED",
    reason="Requires KIVY_LOG_MODE==MIXED to run.",
)
def test_mixed_mode_handlers():
    assert not are_regular_logs_handled()
    assert are_kivy_logger_logs_handled()
    assert not is_stderr_output_handled()



'''
Graphics tests
==============

Testing the simple vertex instructions
'''

import sys
import pytest
from threading import Thread
from kivy.tests.common import GraphicUnitTest, requires_graphics


class BoxShadowTest(GraphicUnitTest):

    def test_create(self):
        from kivy.graphics.boxshadow import BoxShadow
        from kivy.uix.widget import Widget
        from kivy.graphics import Color

        r = self.render

        # with initial arguments
        wid = Widget()
        with wid.canvas:
            Color(1, 0, 0, 1)
            bs = BoxShadow(
                pos=(50, 50),
                size=(150, 150),
                offset=(0, 10),
                spread_radius=(10, -10),
                border_radius=(10, 10, 10, 10),
                blur_radius=80,
            )
        r(wid)

        wid = Widget()
        with wid.canvas:
            Color(0, 1, 0, 1)
            bs = BoxShadow(
                inset=True,
                pos=(50, 50),
                size=(150, 150),
                offset=(0, 10),
                spread_radius=(10, -10),
                border_radius=(10, 10, 10, 10),
                blur_radius=80,
            )
        r(wid)

        # changing properties later
        wid = Widget()
        with wid.canvas:
            Color(0, 0, 1, 1)
            bs = BoxShadow()
        bs.inset = True
        bs.pos = [50, 50]
        bs.size = [150, 150]
        bs.offset = [0, 10]
        bs.spread_radius = [10, -10]
        bs.border_radius = [10, 10, 10, 10]
        bs.blur_radius = 40
        r(wid)

    def test_adjusted_size(self):
        from kivy.graphics.boxshadow import BoxShadow

        raw_size = 150, 150

        bs = BoxShadow()
        bs.pos = 50, 50
        bs.size = raw_size
        bs.blur_radius = 80
        bs.spread_radius = -10, 10

        # The size of the rectangle containing the FBO texture (shadow) needs
        # to be adjusted according to the size of the shadow, otherwise there
        # will be an unwanted cropping behavior.
        adjusted_size = (
            max(
                0,
                raw_size[0] + bs.blur_radius * 3 + bs.spread_radius[0] * 2,
            ),
            max(
                0,
                raw_size[1] + bs.blur_radius * 3 + bs.spread_radius[1] * 2,
            ),
        )

        assert bs.size == adjusted_size

        # Now we will turn on the inset mode, it is expected that
        # there will be no size adjustments.
        bs.inset = True
        assert bs.size == raw_size

        # Now turning off, and reverting back to the default mode.
        bs.inset = False
        assert bs.size == adjusted_size

        # Testing with initial arguments
        bs = BoxShadow(
            inset=True,
            pos=(50, 50),
            size=raw_size,
            blur_radius=80,
            spread_radius=(10, -10)
        )
        adjusted_size = (
            max(
                0,
                raw_size[0] + bs.blur_radius * 3 + bs.spread_radius[0] * 2,
            ),
            max(
                0,
                raw_size[1] + bs.blur_radius * 3 + bs.spread_radius[1] * 2,
            ),
        )

        assert bs.size == raw_size

        # Now turning off, and reverting back to the default mode.
        bs.inset = False
        assert bs.size == adjusted_size

        # Now we will turn on the inset mode, it is expected that
        # there will be no size adjustments.
        bs.inset = True
        assert bs.size == raw_size

    def test_adjusted_pos(self):
        from kivy.graphics.boxshadow import BoxShadow

        raw_pos = 50, 50
        raw_size = 150, 150
        offset = 10, -100

        bs = BoxShadow()
        bs.pos = raw_pos
        bs.size = raw_size
        bs.offset = offset
        bs.blur_radius = 80
        bs.spread_radius = -10, 10

        # If the size of the rectangle containing the FBO texture (shadow)
        # changes, its position will need to be adjusted.
        adjusted_pos = (
            raw_pos[0]
            - bs.blur_radius * 1.5
            - bs.spread_radius[0]
            + bs.offset[0],
            raw_pos[0]
            - bs.blur_radius * 1.5
            - bs.spread_radius[1]
            + bs.offset[1],
        )

        assert bs.pos == adjusted_pos

        # Now we will turn on the inset mode, it is expected that
        # there will be no position adjustments.
        bs.inset = True
        assert bs.pos == raw_pos

        # Now turning off, and reverting back to the default mode.
        bs.inset = False
        assert bs.pos == adjusted_pos

        # Testing with initial arguments
        bs = BoxShadow(
            inset=True,
            pos=raw_pos,
            size=raw_size,
            offset=offset,
            blur_radius=80,
            spread_radius=(10, -10)
        )
        adjusted_pos = (
            raw_pos[0]
            - bs.blur_radius * 1.5
            - bs.spread_radius[0]
            + bs.offset[0],
            raw_pos[0]
            - bs.blur_radius * 1.5
            - bs.spread_radius[1]
            + bs.offset[1],
        )

        assert bs.pos == raw_pos

        # Now turning off, and reverting back to the default mode.
        bs.inset = False
        assert bs.pos == adjusted_pos

        # Now we will turn on the inset mode, it is expected that
        # there will be no position adjustments.
        bs.inset = True
        assert bs.pos == raw_pos

    def test_bounded_properties(self):
        from kivy.graphics.boxshadow import BoxShadow

        bs = BoxShadow()
        bs.pos = 50, 50
        bs.size = 150, 150
        bs.offset = 10, -100
        bs.blur_radius = -80
        bs.spread_radius = -200, -100
        bs.border_radius = 0, 0, 100, 0

        assert bs.size == (0, 0)
        assert bs.blur_radius == 0

        # There is a bug in RoundedRectangle that distorts the texture if the
        # radius value is less than 1. Otherwise, it could be 0.
        assert bs.border_radius == tuple(
            map(
                lambda value: max(1.0, min(value, min(bs.size) / 2)),
                bs.border_radius,
            )
        )

        # Testing with initial arguments
        bs = BoxShadow(
            pos=(50, 50),
            size=(150, 150),
            offset=(10, -100),
            blur_radius=-80,
            spread_radius=(-200, -100),
            border_radius=(0, 0, 100, 0),
        )

        assert bs.size == (0, 0)
        assert bs.blur_radius == 0
        assert bs.border_radius == tuple(
            map(
                lambda value: max(1.0, min(value, min(bs.size) / 2)),
                bs.border_radius,
            )
        )


class VertexInstructionTest(GraphicUnitTest):

    def test_circle(self):
        from kivy.uix.widget import Widget
        from kivy.graphics import Ellipse, Color

        r = self.render

        # basic circle
        wid = Widget()
        with wid.canvas:
            Color(1, 1, 1)
            Ellipse(pos=(100, 100), size=(100, 100))
        r(wid)

        # reduced circle
        wid = Widget()
        with wid.canvas:
            Color(1, 1, 1)
            Ellipse(pos=(100, 100), size=(100, 100), segments=10)
        r(wid)

        # moving circle
        wid = Widget()
        with wid.canvas:
            Color(1, 1, 1)
            self.e = Ellipse(pos=(100, 100), size=(100, 100))
        self.e.pos = (10, 10)
        r(wid)

    def test_ellipse(self):
        from kivy.uix.widget import Widget
        from kivy.graphics import Ellipse, Color

        r = self.render

        # ellipse
        wid = Widget()
        with wid.canvas:
            Color(1, 1, 1)
            self.e = Ellipse(pos=(100, 100), size=(200, 100))
        r(wid)

    def test_point(self):
        from kivy.uix.widget import Widget
        from kivy.graphics import Point, Color

        r = self.render

        # 1 point
        wid = Widget()
        with wid.canvas:
            Color(1, 1, 1)
            Point(points=(10, 10))
        r(wid)

        # 25 points
        wid = Widget()
        with wid.canvas:
            Color(1, 1, 1)
            Point(points=[x * 5 for x in range(50)])
        r(wid)

    def test_point_add(self):
        from kivy.uix.widget import Widget
        from kivy.graphics import Point, Color

        r = self.render

        wid = Widget()
        with wid.canvas:
            Color(1, 1, 1)
            p = Point(pointsize=10)

        p.add_point(10, 10)
        p.add_point(90, 10)
        p.add_point(10, 90)
        p.add_point(50, 50)
        p.add_point(10, 50)
        p.add_point(50, 10)

        r(wid)

    def test_line_rounded_rectangle(self):
        from kivy.uix.widget import Widget
        from kivy.graphics import Line, Color
        r = self.render

        # basic rounded_rectangle
        wid = Widget()
        with wid.canvas:
            Color(1, 1, 1)
            line = Line(
                rounded_rectangle=(100, 100, 100, 100, 10, 20, 30, 40, 100)
            )
        r(wid)
        assert line.rounded_rectangle == (
            100, 100, 100, 100, 10, 20, 30, 40, 100
        )

        # The largest angle allowed is equal to the smallest dimension (width
        # or height) minus the largest angle value between the anterior angle
        # and the posterior angle.
        wid = Widget()
        with wid.canvas:
            Color(1, 1, 1)
            line = Line(
                rounded_rectangle=(
                    100, 100, 100, 100, 100, 20, 10, 30, 100
                )
            )
        r(wid)
        assert line.rounded_rectangle == (
            100, 100, 100, 100, 70, 20, 10, 30, 100
        )

        # Same approach as above
        wid = Widget()
        with wid.canvas:
            Color(1, 1, 1)
            line = Line(
                rounded_rectangle=(
                    100, 100, 100, 100, 100, 25, 100, 50, 100
                )
            )
        r(wid)
        assert line.rounded_rectangle == (
            100, 100, 100, 100, 50, 25, 50, 50, 100
        )

        # A circle should be generated if width and height are equal, and all
        # angles passed are greater than or equal to the smallest dimension.
        wid = Widget()
        with wid.canvas:
            Color(1, 1, 1)
            line = Line(
                rounded_rectangle=(
                    100, 100, 100, 100, 150, 50, 50.001, 51, 100
                )
            )
        r(wid)
        assert line.rounded_rectangle == (
            100, 100, 100, 100, 50, 50, 50, 50, 100
        )

        # Currently the minimum radius should be 1, to avoid rendering issues
        wid = Widget()
        with wid.canvas:
            Color(1, 1, 1)
            line = Line(
                rounded_rectangle=(
                    100, 100, 100, 100, 0, 0, 0, 0, 100
                )
            )
        r(wid)
        assert line.rounded_rectangle == (
            100, 100, 100, 100, 1, 1, 1, 1, 100
        )

        # Angles adjustment + avoid issue if radius is less than 1
        wid = Widget()
        with wid.canvas:
            Color(1, 1, 1)
            line = Line(
                rounded_rectangle=(
                    100, 100, 100, 100, 100, 0, 0, 0, 100
                )
            )
        r(wid)
        assert line.rounded_rectangle == (
            100, 100, 100, 100, 99, 1, 1, 1, 100
        )

    def test_smoothline_rounded_rectangle(self):
        from kivy.uix.widget import Widget
        from kivy.graphics import SmoothLine, Color
        r = self.render

        # If width and/or height < 2px, the figure should not be rendered.
        # This avoids some known SmoothLine rendering issues.
        wid = Widget()
        with wid.canvas:
            Color(1, 1, 1)
            line = SmoothLine(
                rounded_rectangle=(100, 100, 0.5, 1.99, 30, 30, 30, 30, 100)
            )
        r(wid)
        assert line.rounded_rectangle is None


class FBOInstructionTestCase(GraphicUnitTest):

    def test_fbo_pixels(self):
        from kivy.graphics import Fbo, ClearColor, ClearBuffers, Ellipse

        fbo = Fbo(size=(512, 512))
        with fbo:
            ClearColor(0, 0, 0, 1)
            ClearBuffers()
            Ellipse(pos=(100, 100), size=(100, 100))
        fbo.draw()
        data = fbo.pixels
        fbo.texture.save('results.png')


class TransformationsTestCase(GraphicUnitTest):

    def test_identity_creation(self):
        from kivy.graphics import LoadIdentity

        mat = LoadIdentity()
        self.assertTrue(mat.stack)


class CallbackInstructionTest(GraphicUnitTest):

    def test_from_kv(self):
        from textwrap import dedent
        from kivy.lang import Builder

        root = Builder.load_string(dedent("""\
        Widget:
            canvas:
                Callback:
                    callback: lambda __: setattr(self, 'callback_test', 'TEST')
        """))
        r = self.render
        r(root)
        self.assertTrue(root.callback_test == 'TEST')


@pytest.fixture
def widget_verify_thread(request):
    from kivy.uix.widget import Widget
    from kivy.config import Config

    original = Config.get('graphics', 'verify_gl_main_thread')
    Config.set('graphics', 'verify_gl_main_thread', request.param)

    widget = Widget()
    yield widget, request.param

    Config.set('graphics', 'verify_gl_main_thread', original)


@requires_graphics
@pytest.mark.parametrize('widget_verify_thread', ['0', '1'], indirect=True)
def test_graphics_main_thread(widget_verify_thread):
    from kivy.graphics import Color

    widget, verify_thread = widget_verify_thread
    with widget.canvas:
        color = Color()
    color.rgb = .1, .2, .3


@requires_graphics
@pytest.mark.parametrize('widget_verify_thread', ['0', '1'], indirect=True)
def test_create_graphics_second_thread(widget_verify_thread):
    from kivy.graphics import Color
    widget, verify_thread = widget_verify_thread
    exception = None

    def callback():
        nonlocal exception
        try:
            with widget.canvas:
                if verify_thread == '1':
                    with pytest.raises(TypeError):
                        Color()
                else:
                    Color()
        except BaseException as e:
            exception = e, sys.exc_info()[2]
            raise

    thread = Thread(target=callback)
    thread.start()
    thread.join()
    if exception is not None:
        raise exception[0].with_traceback(exception[1])


@requires_graphics
@pytest.mark.parametrize('widget_verify_thread', ['0', '1'], indirect=True)
def test_change_graphics_second_thread(widget_verify_thread):
    from kivy.graphics import Color
    widget, verify_thread = widget_verify_thread
    with widget.canvas:
        color = Color()

    exception = None

    def callback():
        nonlocal exception
        try:
            if verify_thread == '1':
                with pytest.raises(TypeError):
                    color.rgb = .1, .2, .3
            else:
                color.rgb = .1, .2, .3
        except BaseException as e:
            exception = e, sys.exc_info()[2]
            raise

    thread = Thread(target=callback)
    thread.start()
    thread.join()
    if exception is not None:
        raise exception[0].with_traceback(exception[1])



'''
Svg tests
==============

Testing Svg rendering.
'''

from kivy.tests.common import GraphicUnitTest


SIMPLE_SVG = """<?xml version="1.0" standalone="no"?>
<svg width="256" height="256" viewBox="0 0 256 256" version="1.1"
    xmlns="http://www.w3.org/2000/svg">
<rect stroke="blue" stroke-width="4" x="24" y="30" width="92" height="166"
    fill="none" stroke-opacity="0.5" />
</svg>
"""

SCALE_SVG = """<?xml version="1.0" standalone="no"?>
<svg width="256" height="256" viewBox="0 0 256 256" version="1.1"
    xmlns="http://www.w3.org/2000/svg">
<rect stroke="red" stroke-width="4" x="24" y="30" width="10" height="10"
    fill="none" stroke-opacity="0.5" transform="scale(2, 3)"/>
</svg>
"""

ROTATE_SVG = """<?xml version="1.0" standalone="no"?>
<svg width="256" height="256" viewBox="0 0 256 256" version="1.1"
    xmlns="http://www.w3.org/2000/svg">
<rect stroke="green" stroke-width="4" x="24" y="30" width="50" height="100"
    stroke-opacity="0.75" transform="rotate(60 128 128)" />
</svg>
"""


class SvgTest(GraphicUnitTest):

    def test_simple(self):
        import xml.etree.ElementTree as ET
        from kivy.uix.widget import Widget
        from kivy.graphics.svg import Svg

        # create a root widget
        wid = Widget()

        # put some graphics instruction on it
        with wid.canvas:
            svg = Svg()
            svg.set_tree(ET.ElementTree(ET.fromstring(SIMPLE_SVG)))

        # render, and capture it directly
        self.render(wid)

    def test_scale(self):
        import xml.etree.ElementTree as ET
        from kivy.uix.widget import Widget
        from kivy.graphics.svg import Svg

        # create a root widget
        wid = Widget()

        # put some graphics instruction on it
        with wid.canvas:
            svg = Svg()
            svg.set_tree(ET.ElementTree(ET.fromstring(SCALE_SVG)))

        # render, and capture it directly
        self.render(wid)

    def test_rotate(self):
        import xml.etree.ElementTree as ET
        from kivy.uix.widget import Widget
        from kivy.graphics.svg import Svg

        # create a root widget
        wid = Widget()

        # put some graphics instruction on it
        with wid.canvas:
            svg = Svg()
            svg.set_tree(ET.ElementTree(ET.fromstring(ROTATE_SVG)))

        # render, and capture it directly
        self.render(wid)
